@@ -24,9 +24,30 @@ from .. import mlog
 from .. import compilers
 from ..build import BuildTarget
 from ..compilers import CompilerArgs
-from ..mesonlib import MesonException, File
+from ..mesonlib import MesonException, File, get_meson_script
 from ..environment import Environment
 
+def autodetect_vs_version(build):
+    vs_version = os.getenv('VisualStudioVersion', None)
+    vs_install_dir = os.getenv('VSINSTALLDIR', None)
+    if not vs_version and not vs_install_dir:
+        raise MesonException('Could not detect Visual Studio: VisualStudioVersion and VSINSTALLDIR are unset!\n'
+                             'Are we inside a Visual Studio build environment? '
+                             'You can also try specifying the exact backend to use.')
+    # VisualStudioVersion is set since Visual Studio 12.0, but sometimes
+    # vcvarsall.bat doesn't set it, so also use VSINSTALLDIR
+    if vs_version == '14.0' or 'Visual Studio 14' in vs_install_dir:
+        from mesonbuild.backend.vs2015backend import Vs2015Backend
+        return Vs2015Backend(build)
+    if vs_version == '15.0' or 'Visual Studio 17' in vs_install_dir or \
+       'Visual Studio\\2017' in vs_install_dir:
+        from mesonbuild.backend.vs2017backend import Vs2017Backend
+        return Vs2017Backend(build)
+    if 'Visual Studio 10.0' in vs_install_dir:
+        return Vs2010Backend(build)
+    raise MesonException('Could not detect Visual Studio using VisualStudioVersion: {!r} or VSINSTALLDIR: {!r}!\n'
+                         'Please specify the exact backend to use.'.format(vs_version, vs_install_dir))
+
 def split_o_flags_args(args):
     """
     Splits any /O args and returns them. Does not take care of flags overriding
@@ -62,8 +83,9 @@ class Vs2010Backend(backends.Backend):
         self.sources_conflicts = {}
         self.platform_toolset = None
         self.vs_version = '2010'
+        self.windows_target_platform_version = None
 
-    def object_filename_from_source(self, target, source):
+    def object_filename_from_source(self, target, source, is_unity=False):
         basename = os.path.basename(source.fname)
         filename_without_extension = '.'.join(basename.split('.')[:-1])
         if basename in self.sources_conflicts[target.get_id()]:
@@ -93,12 +115,10 @@ class Vs2010Backend(backends.Backend):
 
     def generate_custom_generator_commands(self, target, parent_node):
         generator_output_files = []
-        commands = []
-        inputs = []
-        outputs = []
         custom_target_include_dirs = []
         custom_target_output_files = []
         target_private_dir = self.relpath(self.get_target_private_dir(target), self.get_target_dir(target))
+        source_target_dir = self.get_target_source_dir(target)
         down = self.target_to_build_root(target)
         for genlist in target.get_generated_sources():
             if isinstance(genlist, build.CustomTarget):
@@ -116,6 +136,7 @@ class Vs2010Backend(backends.Backend):
                 outfilelist = genlist.get_outputs()
                 exe_arr = self.exe_object_to_cmd_array(exe)
                 base_args = generator.get_arglist()
+                idgroup = ET.SubElement(parent_node, 'ItemGroup')
                 for i in range(len(infilelist)):
                     if len(infilelist) == len(outfilelist):
                         sole_output = os.path.join(target_private_dir, outfilelist[i])
@@ -129,21 +150,17 @@ class Vs2010Backend(backends.Backend):
                     args = [x.replace("@INPUT@", infilename).replace('@OUTPUT@', sole_output)
                             for x in base_args]
                     args = self.replace_outputs(args, target_private_dir, outfiles_rel)
-                    args = [x.replace("@SOURCE_DIR@", self.environment.get_source_dir()).replace("@BUILD_DIR@", target_private_dir)
+                    args = [x.replace("@SOURCE_DIR@", self.environment.get_source_dir())
+                             .replace("@BUILD_DIR@", target_private_dir)
                             for x in args]
-                    fullcmd = exe_arr + self.replace_extra_args(args, genlist)
-                    commands.append(' '.join(self.special_quote(fullcmd)))
-                    inputs.append(infilename)
-                    outputs.extend(outfiles)
-        if len(commands) > 0:
-            idgroup = ET.SubElement(parent_node, 'ItemDefinitionGroup')
-            cbs = ET.SubElement(idgroup, 'CustomBuildStep')
-            ET.SubElement(cbs, 'Command').text = '\r\n'.join(commands)
-            ET.SubElement(cbs, 'Inputs').text = ";".join(inputs)
-            ET.SubElement(cbs, 'Outputs').text = ';'.join(outputs)
-            ET.SubElement(cbs, 'Message').text = 'Generating custom sources.'
-            pg = ET.SubElement(parent_node, 'PropertyGroup')
-            ET.SubElement(pg, 'CustomBuildBeforeTargets').text = 'ClCompile'
+                    args = [x.replace("@CURRENT_SOURCE_DIR@", source_target_dir) for x in args]
+                    args = [x.replace("@SOURCE_ROOT@", self.environment.get_source_dir())
+                             .replace("@BUILD_ROOT@", self.environment.get_build_dir())
+                            for x in args]
+                    cmd = exe_arr + self.replace_extra_args(args, genlist)
+                    cbs = ET.SubElement(idgroup, 'CustomBuild', Include=infilename)
+                    ET.SubElement(cbs, 'Command').text = ' '.join(self.quote_arguments(cmd))
+                    ET.SubElement(cbs, 'Outputs').text = ';'.join(outfiles)
         return generator_output_files, custom_target_output_files, custom_target_include_dirs
 
     def generate(self, interp):
@@ -205,10 +222,11 @@ class Vs2010Backend(backends.Backend):
                 for d in [target.command] + target.args:
                     if isinstance(d, (build.BuildTarget, build.CustomTarget)):
                         all_deps[d.get_id()] = d
-            # BuildTarget
-            else:
+            elif isinstance(target, build.BuildTarget):
                 for ldep in target.link_targets:
                     all_deps[ldep.get_id()] = ldep
+                for ldep in target.link_whole_targets:
+                    all_deps[ldep.get_id()] = ldep
                 for obj_id, objdep in self.get_obj_target_deps(target.objects):
                     all_deps[obj_id] = objdep
                 for gendep in target.get_generated_sources():
@@ -218,6 +236,8 @@ class Vs2010Backend(backends.Backend):
                         gen_exe = gendep.generator.get_exe()
                         if isinstance(gen_exe, build.Executable):
                             all_deps[gen_exe.get_id()] = gen_exe
+            else:
+                raise MesonException('Unknown target type for target %s' % target)
         if not t or not recursive:
             return all_deps
         ret = self.get_target_deps(all_deps, recursive)
@@ -340,8 +360,8 @@ class Vs2010Backend(backends.Backend):
         directories = os.path.normpath(target.subdir).split(os.sep)
         return os.sep.join(['..'] * len(directories))
 
-    def special_quote(self, arr):
-        return ['&quot;%s&quot;' % i for i in arr]
+    def quote_arguments(self, arr):
+        return ['"%s"' % i for i in arr]
 
     def create_basic_crap(self, target):
         project_name = target.name
@@ -357,13 +377,15 @@ class Vs2010Backend(backends.Backend):
         pl.text = self.platform
         globalgroup = ET.SubElement(root, 'PropertyGroup', Label='Globals')
         guidelem = ET.SubElement(globalgroup, 'ProjectGuid')
-        guidelem.text = self.environment.coredata.test_guid
+        guidelem.text = '{%s}' % self.environment.coredata.test_guid
         kw = ET.SubElement(globalgroup, 'Keyword')
         kw.text = self.platform + 'Proj'
         p = ET.SubElement(globalgroup, 'Platform')
         p.text = self.platform
         pname = ET.SubElement(globalgroup, 'ProjectName')
         pname.text = project_name
+        if self.windows_target_platform_version:
+            ET.SubElement(globalgroup, 'WindowsTargetPlatformVersion').text = self.windows_target_platform_version
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.Default.props')
         type_config = ET.SubElement(root, 'PropertyGroup', Label='Configuration')
         ET.SubElement(type_config, 'ConfigurationType')
@@ -389,21 +411,25 @@ class Vs2010Backend(backends.Backend):
         customstep = ET.SubElement(action, 'PostBuildEvent')
         cmd_raw = [target.command] + target.args
         cmd = [sys.executable, os.path.join(self.environment.get_script_dir(), 'commandrunner.py'),
-               self.environment.get_build_dir(), self.environment.get_source_dir(),
-               self.get_target_dir(target)]
+               self.environment.get_build_dir(),
+               self.environment.get_source_dir(),
+               self.get_target_dir(target),
+               get_meson_script(self.environment, 'mesonintrospect')]
         for i in cmd_raw:
             if isinstance(i, build.BuildTarget):
                 cmd.append(os.path.join(self.environment.get_build_dir(), self.get_target_filename(i)))
             elif isinstance(i, dependencies.ExternalProgram):
                 cmd += i.get_command()
+            elif isinstance(i, File):
+                relfname = i.rel_to_builddir(self.build_to_src)
+                cmd.append(os.path.join(self.environment.get_build_dir(), relfname))
             else:
                 cmd.append(i)
         cmd_templ = '''"%s" ''' * len(cmd)
         ET.SubElement(customstep, 'Command').text = cmd_templ % tuple(cmd)
         ET.SubElement(customstep, 'Message').text = 'Running custom command.'
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.targets')
-        tree = ET.ElementTree(root)
-        tree.write(ofname, encoding='utf-8', xml_declaration=True)
+        self._prettyprint_vcxproj_xml(ET.ElementTree(root), ofname)
 
     def gen_custom_target_vcxproj(self, target, ofname, guid):
         root = self.create_basic_crap(target)
@@ -413,13 +439,22 @@ class Vs2010Backend(backends.Backend):
         # from the target dir, not the build root.
         target.absolute_paths = True
         (srcs, ofilenames, cmd) = self.eval_custom_target_command(target, True)
-        cmd_templ = '''"%s" ''' * len(cmd)
-        ET.SubElement(customstep, 'Command').text = cmd_templ % tuple(cmd)
+        depend_files = self.get_custom_target_depend_files(target, True)
+        # Always use a wrapper because MSBuild eats random characters when
+        # there are many arguments.
+        tdir_abs = os.path.join(self.environment.get_build_dir(), self.get_target_dir(target))
+        exe_data = self.serialize_executable(target.command[0], cmd[1:],
+                                             # All targets run from the target dir
+                                             tdir_abs,
+                                             capture=ofilenames[0] if target.capture else None)
+        wrapper_cmd = [sys.executable, self.environment.get_build_command(),
+                       '--internal', 'exe', exe_data]
+        ET.SubElement(customstep, 'Command').text = ' '.join(self.quote_arguments(wrapper_cmd))
         ET.SubElement(customstep, 'Outputs').text = ';'.join(ofilenames)
-        ET.SubElement(customstep, 'Inputs').text = ';'.join(srcs)
+        ET.SubElement(customstep, 'Inputs').text = ';'.join([exe_data] + srcs + depend_files)
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.targets')
-        tree = ET.ElementTree(root)
-        tree.write(ofname, encoding='utf-8', xml_declaration=True)
+        self.generate_custom_generator_commands(target, root)
+        self._prettyprint_vcxproj_xml(ET.ElementTree(root), ofname)
 
     @classmethod
     def lang_from_source_file(cls, src):
@@ -467,14 +502,14 @@ class Vs2010Backend(backends.Backend):
 
     @staticmethod
     def has_objects(objects, additional_objects, generated_objects):
-        # Ignore generated objects, those are automatically used by MSBuild for VS2010, because they are part of
-        # the CustomBuildStep Outputs.
+        # Ignore generated objects, those are automatically used by MSBuild because they are part of
+        # the CustomBuild Outputs.
         return len(objects) + len(additional_objects) > 0
 
     @staticmethod
     def add_generated_objects(node, generated_objects):
-        # Do not add generated objects to project file. Those are automatically used by MSBuild for VS2010, because
-        # they are part of the CustomBuildStep Outputs.
+        # Do not add generated objects to project file. Those are automatically used by MSBuild, because
+        # they are part of the CustomBuild Outputs.
         return
 
     @staticmethod
@@ -545,6 +580,13 @@ class Vs2010Backend(backends.Backend):
                     return c
         raise MesonException('Could not find a C or C++ compiler. MSVC can only build C/C++ projects.')
 
+    def _prettyprint_vcxproj_xml(self, tree, ofname):
+        tree.write(ofname, encoding='utf-8', xml_declaration=True)
+        # ElementTree can not do prettyprinting so do it manually
+        doc = xml.dom.minidom.parse(ofname)
+        with open(ofname, 'w') as of:
+            of.write(doc.toprettyxml())
+
     def gen_vcxproj(self, target, ofname, guid):
         mlog.debug('Generating vcxproj %s.' % target.name)
         entrypoint = 'WinMainCRTStartup'
@@ -572,6 +614,8 @@ class Vs2010Backend(backends.Backend):
         # Prefix to use to access the source tree's subdir from the vcxproj dir
         proj_to_src_dir = os.path.join(proj_to_src_root, target.subdir)
         (sources, headers, objects, languages) = self.split_sources(target.sources)
+        if self.is_unity(target):
+            sources = self.generate_unity_files(target, sources)
         compiler = self._get_cl_compiler(target)
         buildtype_args = compiler.get_buildtype_args(self.buildtype)
         buildtype_link_args = compiler.get_buildtype_linker_args(self.buildtype)
@@ -590,7 +634,7 @@ class Vs2010Backend(backends.Backend):
         # Globals
         globalgroup = ET.SubElement(root, 'PropertyGroup', Label='Globals')
         guidelem = ET.SubElement(globalgroup, 'ProjectGuid')
-        guidelem.text = guid
+        guidelem.text = '{%s}' % guid
         kw = ET.SubElement(globalgroup, 'Keyword')
         kw.text = self.platform + 'Proj'
         ns = ET.SubElement(globalgroup, 'RootNamespace')
@@ -599,6 +643,8 @@ class Vs2010Backend(backends.Backend):
         p.text = self.platform
         pname = ET.SubElement(globalgroup, 'ProjectName')
         pname.text = project_name
+        if self.windows_target_platform_version:
+            ET.SubElement(globalgroup, 'WindowsTargetPlatformVersion').text = self.windows_target_platform_version
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.Default.props')
         # Start configuration
         type_config = ET.SubElement(root, 'PropertyGroup', Label='Configuration')
@@ -658,9 +704,6 @@ class Vs2010Backend(backends.Backend):
             ET.SubElement(type_config, 'Optimization').text = 'MinSpace'
         elif '/Od' in o_flags:
             ET.SubElement(type_config, 'Optimization').text = 'Disabled'
-        # Warning level
-        warning_level = self.environment.coredata.get_builtin_option('warning_level')
-        ET.SubElement(type_config, 'WarningLevel').text = 'Level' + warning_level
         # End configuration
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.props')
         generated_files, custom_target_output_files, generated_files_include_dirs = self.generate_custom_generator_commands(target, root)
@@ -711,11 +754,12 @@ class Vs2010Backend(backends.Backend):
         for l, args in self.build.global_args.items():
             if l in file_args:
                 file_args[l] += args
-        # Compile args added from the env: CFLAGS/CXXFLAGS, etc. We want these
-        # to override all the defaults, but not the per-target compile args.
-        for l, args in self.environment.coredata.external_args.items():
-            if l in file_args:
-                file_args[l] += args
+        if not target.is_cross:
+            # Compile args added from the env: CFLAGS/CXXFLAGS, etc. We want these
+            # to override all the defaults, but not the per-target compile args.
+            for l, args in self.environment.coredata.external_args.items():
+                if l in file_args:
+                    file_args[l] += args
         for args in file_args.values():
             # This is where Visual Studio will insert target_args, target_defines,
             # etc, which are added later from external deps (see below).
@@ -724,12 +768,15 @@ class Vs2010Backend(backends.Backend):
             # and from `include_directories:` of internal deps of the target.
             #
             # Target include dirs should override internal deps include dirs.
+            # This is handled in BuildTarget.process_kwargs()
             #
             # Include dirs from internal deps should override include dirs from
-            # external deps.
+            # external deps and must maintain the order in which they are
+            # specified. Hence, we must reverse so that the order is preserved.
+            #
             # These are per-target, but we still add them as per-file because we
             # need them to be looked in first.
-            for d in target.get_include_dirs():
+            for d in reversed(target.get_include_dirs()):
                 for i in d.get_incdirs():
                     curdir = os.path.join(d.get_curdir(), i)
                     args.append('-I' + self.relpath(curdir, target.subdir)) # build dir
@@ -746,9 +793,13 @@ class Vs2010Backend(backends.Backend):
         # target private dir, target build dir, generated sources include dirs,
         # target source dir
         for args in file_args.values():
-            t_inc_dirs = ['.', self.relpath(self.get_target_private_dir(target),
-                                            self.get_target_dir(target))]
-            t_inc_dirs += generated_files_include_dirs + [proj_to_src_dir]
+            t_inc_dirs = [self.relpath(self.get_target_private_dir(target),
+                                       self.get_target_dir(target))]
+            if target.implicit_include_directories:
+                t_inc_dirs += ['.']
+            t_inc_dirs += generated_files_include_dirs
+            if target.implicit_include_directories:
+                t_inc_dirs += [proj_to_src_dir]
             args += ['-I' + arg for arg in t_inc_dirs]
 
         # Split preprocessor defines and include directories out of the list of
@@ -779,7 +830,7 @@ class Vs2010Backend(backends.Backend):
 
         # Split compile args needed to find external dependencies
         # Link args are added while generating the link command
-        for d in target.get_external_deps():
+        for d in reversed(target.get_external_deps()):
             # Cflags required by external deps might have UNIX-specific flags,
             # so filter them out if needed
             d_compile_args = compiler.unix_args_to_native(d.get_compile_args())
@@ -810,13 +861,16 @@ class Vs2010Backend(backends.Backend):
         ET.SubElement(clconf, 'MinimalRebuild').text = 'true'
         ET.SubElement(clconf, 'FunctionLevelLinking').text = 'true'
         pch_node = ET.SubElement(clconf, 'PrecompiledHeader')
-        if self.environment.coredata.get_builtin_option('werror'):
+        # Warning level
+        warning_level = self.get_option_for_target('warning_level', target)
+        ET.SubElement(clconf, 'WarningLevel').text = 'Level' + str(1 + int(warning_level))
+        if self.get_option_for_target('werror', target):
             ET.SubElement(clconf, 'TreatWarningAsError').text = 'true'
         # Note: SuppressStartupBanner is /NOLOGO and is 'true' by default
         pch_sources = {}
         for lang in ['c', 'cpp']:
             pch = target.get_pch(lang)
-            if len(pch) == 0:
+            if not pch:
                 continue
             pch_node.text = 'Use'
             pch_sources[lang] = [pch[0], pch[1], lang]
@@ -850,18 +904,21 @@ class Vs2010Backend(backends.Backend):
             # Add link args added using add_global_link_arguments()
             # These override per-project link arguments
             extra_link_args += self.build.get_global_link_args(compiler)
-            # Link args added from the env: LDFLAGS. We want these to
-            # override all the defaults but not the per-target link args.
-            extra_link_args += self.environment.coredata.external_link_args[compiler.get_language()]
+            if not target.is_cross:
+                # Link args added from the env: LDFLAGS. We want these to
+                # override all the defaults but not the per-target link args.
+                extra_link_args += self.environment.coredata.external_link_args[compiler.get_language()]
             # Only non-static built targets need link args and link dependencies
             extra_link_args += target.link_args
             # External deps must be last because target link libraries may depend on them.
             for dep in target.get_external_deps():
-                extra_link_args += dep.get_link_args()
+                # Extend without reordering or de-dup to preserve `-L -l` sets
+                # https://github.com/mesonbuild/meson/issues/1718
+                extra_link_args.extend_direct(dep.get_link_args())
             for d in target.get_dependencies():
                 if isinstance(d, build.StaticLibrary):
                     for dep in d.get_external_deps():
-                        extra_link_args += dep.get_link_args()
+                        extra_link_args.extend_direct(dep.get_link_args())
         # Add link args for c_* or cpp_* build options. Currently this only
         # adds c_winlibs and cpp_winlibs when building for Windows. This needs
         # to be after all internal and external libraries so that unresolved
@@ -880,6 +937,8 @@ class Vs2010Backend(backends.Backend):
         for t in target.get_dependencies():
             lobj = self.build.targets[t.get_id()]
             linkname = os.path.join(down, self.get_target_filename_for_linking(lobj))
+            if t in target.link_whole_targets:
+                linkname = compiler.get_link_whole_for(linkname)[0]
             additional_links.append(linkname)
         for lib in self.get_custom_target_provided_libraries(target):
             additional_links.append(self.relpath(lib, self.get_target_dir(target)))
@@ -888,7 +947,7 @@ class Vs2010Backend(backends.Backend):
             assert(isinstance(o, str))
             additional_objects.append(o)
         for o in custom_objs:
-            additional_objects.append(self.relpath(o, self.get_target_dir(target)))
+            additional_objects.append(o)
         if len(additional_links) > 0:
             additional_links.append('%(AdditionalDependencies)')
             ET.SubElement(link, 'AdditionalDependencies').text = ';'.join(additional_links)
@@ -896,10 +955,12 @@ class Vs2010Backend(backends.Backend):
         ofile.text = '$(OutDir)%s' % target.get_filename()
         subsys = ET.SubElement(link, 'SubSystem')
         subsys.text = subsystem
-        if isinstance(target, build.SharedLibrary):
+        if (isinstance(target, build.SharedLibrary) or
+            isinstance(target, build.Executable)) and target.get_import_filename():
             # DLLs built with MSVC always have an import library except when
             # they're data-only DLLs, but we don't support those yet.
             ET.SubElement(link, 'ImportLibrary').text = target.get_import_filename()
+        if isinstance(target, build.SharedLibrary):
             # Add module definitions file, if provided
             if target.vs_module_defs:
                 relpath = os.path.join(down, target.vs_module_defs.rel_to_builddir(self.build_to_src))
@@ -929,7 +990,7 @@ class Vs2010Backend(backends.Backend):
             for h in gen_hdrs:
                 ET.SubElement(inc_hdrs, 'CLInclude', Include=h)
             for h in target.extra_files:
-                relpath = os.path.join(proj_to_src_dir, h)
+                relpath = os.path.join(down, h.rel_to_builddir(self.build_to_src))
                 ET.SubElement(inc_hdrs, 'CLInclude', Include=relpath)
 
         if len(sources) + len(gen_src) + len(pch_sources) > 0:
@@ -982,19 +1043,7 @@ class Vs2010Backend(backends.Backend):
         ig = ET.SubElement(root, 'ItemGroup')
         pref = ET.SubElement(ig, 'ProjectReference', Include=os.path.join(self.environment.get_build_dir(), 'REGEN.vcxproj'))
         ET.SubElement(pref, 'Project').text = self.environment.coredata.regen_guid
-        tree = ET.ElementTree(root)
-        tree.write(ofname, encoding='utf-8', xml_declaration=True)
-        # ElementTree can not do prettyprinting so do it manually
-        doc = xml.dom.minidom.parse(ofname)
-        with open(ofname, 'w') as of:
-            of.write(doc.toprettyxml())
-        # World of horror! Python insists on not quoting quotes and
-        # fixing the escaped &quot; into &amp;quot; whereas MSVS
-        # requires quoted but not fixed elements. Enter horrible hack.
-        with open(ofname, 'r') as of:
-            txt = of.read()
-        with open(ofname, 'w') as of:
-            of.write(txt.replace('&amp;quot;', '&quot;'))
+        self._prettyprint_vcxproj_xml(ET.ElementTree(root), ofname)
 
     def gen_regenproj(self, project_name, ofname):
         root = ET.Element('Project', {'DefaultTargets': 'Build',
@@ -1009,13 +1058,15 @@ class Vs2010Backend(backends.Backend):
         pl.text = self.platform
         globalgroup = ET.SubElement(root, 'PropertyGroup', Label='Globals')
         guidelem = ET.SubElement(globalgroup, 'ProjectGuid')
-        guidelem.text = self.environment.coredata.test_guid
+        guidelem.text = '{%s}' % self.environment.coredata.test_guid
         kw = ET.SubElement(globalgroup, 'Keyword')
         kw.text = self.platform + 'Proj'
         p = ET.SubElement(globalgroup, 'Platform')
         p.text = self.platform
         pname = ET.SubElement(globalgroup, 'ProjectName')
         pname.text = project_name
+        if self.windows_target_platform_version:
+            ET.SubElement(globalgroup, 'WindowsTargetPlatformVersion').text = self.windows_target_platform_version
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.Default.props')
         type_config = ET.SubElement(root, 'PropertyGroup', Label='Configuration')
         ET.SubElement(type_config, 'ConfigurationType').text = "Utility"
@@ -1071,8 +1122,7 @@ if %%errorlevel%% neq 0 goto :VCEnd'''
         ET.SubElement(custombuild, 'AdditionalInputs').text = ';'.join(deps)
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.targets')
         ET.SubElement(root, 'ImportGroup', Label='ExtensionTargets')
-        tree = ET.ElementTree(root)
-        tree.write(ofname, encoding='utf-8', xml_declaration=True)
+        self._prettyprint_vcxproj_xml(ET.ElementTree(root), ofname)
 
     def gen_testproj(self, target_name, ofname):
         project_name = target_name
@@ -1088,13 +1138,15 @@ if %%errorlevel%% neq 0 goto :VCEnd'''
         pl.text = self.platform
         globalgroup = ET.SubElement(root, 'PropertyGroup', Label='Globals')
         guidelem = ET.SubElement(globalgroup, 'ProjectGuid')
-        guidelem.text = self.environment.coredata.test_guid
+        guidelem.text = '{%s}' % self.environment.coredata.test_guid
         kw = ET.SubElement(globalgroup, 'Keyword')
         kw.text = self.platform + 'Proj'
         p = ET.SubElement(globalgroup, 'Platform')
         p.text = self.platform
         pname = ET.SubElement(globalgroup, 'ProjectName')
         pname.text = project_name
+        if self.windows_target_platform_version:
+            ET.SubElement(globalgroup, 'WindowsTargetPlatformVersion').text = self.windows_target_platform_version
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.Default.props')
         type_config = ET.SubElement(root, 'PropertyGroup', Label='Configuration')
         ET.SubElement(type_config, 'ConfigurationType')
@@ -1124,11 +1176,9 @@ if %%errorlevel%% neq 0 goto :VCEnd'''
         postbuild = ET.SubElement(action, 'PostBuildEvent')
         ET.SubElement(postbuild, 'Message')
         # FIXME: No benchmarks?
-        meson_py = self.environment.get_build_command()
-        (base, ext) = os.path.splitext(meson_py)
-        mesontest_py = base + 'test' + ext
         test_command = [sys.executable,
-                        mesontest_py,
+                        self.environment.get_build_command(),
+                        'test',
                         '--no-rebuild']
         if not self.environment.coredata.get_builtin_option('stdsplit'):
             test_command += ['--no-stdsplit']
@@ -1143,12 +1193,8 @@ endlocal & call :cmErrorLevel %%errorlevel%% & goto :cmDone
 exit /b %%1
 :cmDone
 if %%errorlevel%% neq 0 goto :VCEnd'''
-        self.serialise_tests()
+        self.serialize_tests()
         ET.SubElement(postbuild, 'Command').text =\
             cmd_templ % ('" "'.join(test_command))
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.targets')
-        tree = ET.ElementTree(root)
-        tree.write(ofname, encoding='utf-8', xml_declaration=True)
-        # ElementTree can not do prettyprinting so do it manually
-        # doc = xml.dom.minidom.parse(ofname)
-        # open(ofname, 'w').write(doc.toprettyxml())
+        self._prettyprint_vcxproj_xml(ET.ElementTree(root), ofname)
