@@ -26,8 +26,9 @@ from .. import mesonlib
 from ..mesonlib import MesonException, Popen_safe, flatten, version_compare_many
 
 
-# This must be defined in this file to avoid cyclical references.
+# These must be defined in this file to avoid cyclical references.
 packages = {}
+_packages_accept_language = set()
 
 
 class DependencyException(MesonException):
@@ -43,6 +44,8 @@ class DependencyMethods(Enum):
     SYSTEM = 'system'
     # Detect using sdl2-config
     SDLCONFIG = 'sdlconfig'
+    # Detect using pcap-config
+    PCAPCONFIG = 'pcap-config'
     # This is only supported on OSX - search the frameworks directory by name.
     EXTRAFRAMEWORK = 'extraframework'
     # Detect using the sysconfig module.
@@ -528,7 +531,10 @@ class ExternalLibrary(ExternalDependency):
         C-like code. Note that C++ libraries *can* be linked with C code with
         a C++ linker (and vice-versa).
         '''
-        if self.language == 'vala' and language != 'vala':
+        # Using a vala library in a non-vala target, or a non-vala library in a vala target
+        # XXX: This should be extended to other non-C linkers such as Rust
+        if (self.language == 'vala' and language != 'vala') or \
+           (language == 'vala' and self.language != 'vala'):
             return []
         return self.link_args
 
@@ -548,7 +554,7 @@ class ExtraFrameworkDependency(ExternalDependency):
     def detect(self, name, path):
         lname = name.lower()
         if path is None:
-            paths = ['/Library/Frameworks']
+            paths = ['/System/Library/Frameworks', '/Library/Frameworks']
         else:
             paths = [path]
         for p in paths:
@@ -607,10 +613,15 @@ def find_external_dependency(name, env, kwargs):
         raise DependencyException('Keyword "method" must be a string.')
     lname = name.lower()
     if lname in packages:
+        if lname not in _packages_accept_language and 'language' in kwargs:
+            raise DependencyException('%s dependency does not accept "language" keyword argument' % (lname, ))
         dep = packages[lname](env, kwargs)
         if required and not dep.found():
             raise DependencyException('Dependency "%s" not found' % name)
         return dep
+    if 'language' in kwargs:
+        # Remove check when PkgConfigDependency supports language.
+        raise DependencyException('%s dependency does not accept "language" keyword argument' % (lname, ))
     pkg_exc = None
     pkgdep = None
     try:
