@@ -17,10 +17,11 @@ import time, datetime
 import os.path
 from . import environment, interpreter, mesonlib
 from . import build
+from . import mconf, mintro, mtest, rewriter
 import platform
 from . import mlog, coredata
 from .mesonlib import MesonException
-from .wrap import WrapMode
+from .wrap import WrapMode, wraptool
 
 
 parser = argparse.ArgumentParser()
@@ -180,20 +181,26 @@ If you want to change option values, use the mesonconf tool instead.'''
         intr.run()
         coredata_mtime = time.time()
         g.generate(intr)
-        g.run_postconf_scripts()
         dumpfile = os.path.join(env.get_scratch_dir(), 'build.dat')
         with open(dumpfile, 'wb') as f:
             pickle.dump(b, f)
-        # Write this last since we use the existence of this file to check if
-        # we generated the build file successfully, so we don't want an error
-        # that pops up during generation, post-conf scripts, etc to cause us to
+        # Write this as late as possible since we use the existence of this
+        # file to check if we generated the build file successfully, so we
+        # don't want an error that pops up during generation, etc to cause us to
         # incorrectly signal a successful meson run which will cause an error
         # about an already-configured build directory when the user tries again.
         #
         # However, we set the mtime to an earlier value to ensure that doing an
         # mtime comparison between the coredata dump and other build files
         # shows the build files to be newer, not older.
-        env.dump_coredata(coredata_mtime)
+        cdf = env.dump_coredata(coredata_mtime)
+        # Post-conf scripts must be run after writing coredata or else introspection fails.
+        try:
+            g.run_postconf_scripts()
+        except:
+            os.unlink(cdf)
+            raise
+
 
 def run_script_command(args):
     cmdname = args[0]
@@ -256,12 +263,33 @@ def run_script_command(args):
         raise MesonException('Unknown internal command {}.'.format(cmdname))
     return cmdfunc(cmdargs)
 
-def run(mainfile, args):
-    if sys.version_info < (3, 3):
-        print('Meson works correctly only with python 3.3+.')
+def run(args, mainfile=None):
+    if sys.version_info < (3, 4):
+        print('Meson works correctly only with python 3.4+.')
         print('You have python %s.' % sys.version)
         print('Please update your environment')
         return 1
+    if len(args) > 0:
+        # First check if we want to run a subcommand.
+        cmd_name = args[0]
+        remaining_args = args[1:]
+        if cmd_name == 'test':
+            return mtest.run(remaining_args)
+        elif cmd_name == 'setup':
+            args = remaining_args
+            # FALLTHROUGH like it's 1972.
+        elif cmd_name == 'introspect':
+            return mintro.run(remaining_args)
+        elif cmd_name == 'test':
+            return mtest.run(remaining_args)
+        elif cmd_name == 'rewrite':
+            return rewriter.run(remaining_args)
+        elif cmd_name == 'configure':
+            return mconf.run(remaining_args)
+        elif cmd_name == 'wrap':
+            return wraptool.run(remaining_args)
+
+    # No special command? Do the basic setup/reconf.
     if len(args) >= 2 and args[0] == '--internal':
         if args[1] != 'regenerate':
             script = args[1]
@@ -275,6 +303,7 @@ def run(mainfile, args):
         handshake = True
     else:
         handshake = False
+
     args = mesonlib.expand_arguments(args)
     options = parser.parse_args(args)
     args = options.directories
@@ -296,6 +325,8 @@ def run(mainfile, args):
         else:
             dir2 = '.'
     try:
+        if mainfile is None:
+            sys.exit('I iz broken. Sorry.')
         app = MesonApp(dir1, dir2, mainfile, handshake, options, sys.argv)
     except Exception as e:
         # Log directory does not exist, so just print
