@@ -14,9 +14,11 @@
 
 from . import ModuleReturnValue
 from . import ExtensionModule
+from . import noKwargs
 
 class TestModule(ExtensionModule):
 
+    @noKwargs
     def print_hello(self, state, args, kwargs):
         print('Hello from a Meson module')
         rv = ModuleReturnValue(None, [])
