@@ -18,17 +18,18 @@ from . import mesonlib
 import os, re
 
 forbidden_option_names = coredata.get_builtin_options()
-forbidden_prefixes = {'c_': True,
-                      'cpp_': True,
-                      'd_': True,
-                      'rust_': True,
-                      'fortran_': True,
-                      'objc_': True,
-                      'objcpp_': True,
-                      'vala_': True,
-                      'csharp_': True,
-                      'swift_': True,
-                      'b_': True,
+forbidden_prefixes = {'c_',
+                      'cpp_',
+                      'd_',
+                      'rust_',
+                      'fortran_',
+                      'objc_',
+                      'objcpp_',
+                      'vala_',
+                      'csharp_',
+                      'swift_',
+                      'b_',
+                      'backend_',
                       }
 
 def is_invalid_name(name):
@@ -75,15 +76,16 @@ class OptionInterpreter:
         self.cmd_line_options = {}
         for o in command_line_options:
             if self.subproject != '': # Strip the beginning.
+                # Ignore options that aren't for this subproject
                 if not o.startswith(self.sbprefix):
                     continue
-            else:
-                if ':' in o:
-                    continue
             try:
                 (key, value) = o.split('=', 1)
             except ValueError:
                 raise OptionException('Option {!r} must have a value separated by equals sign.'.format(o))
+            # Ignore subproject options if not fetching subproject options
+            if self.subproject == '' and ':' in key:
+                continue
             self.cmd_line_options[key] = value
 
     def process(self, option_file):
